;----------------------------------------------------------------------------
;Constants

SPBRG_VAL	EQU	 04Fh		;set baud rate 31250 for 40Mhz clock

TX_BUF_LEN	EQU	010h		;length of transmit circular buffer
RX_BUF_LEN	EQU	TX_BUF_LEN	;length of receive circular buffer

;----------------------------------------------------------------------------
;Bit Definitions

TxBufFull	EQU	0		;bit indicates Tx buffer is full
TxBufEmpty	EQU	1		;bit indicates Tx buffer is empty
RxBufFull	EQU	2		;bit indicates Rx buffer is full
RxBufEmpty	EQU	3		;bit indicates Rx buffer is empty

		CBLOCK	0x010
		MidiInByte		; Rcv'd Byte
		MidiByte0		; status
		MidiByte1		; data1
		MidiByte2		; data2
		MidiCurData		; current data byte index
		MidiNumData		; expected data bytes
		MidiTestByte	; for testing only
		ENDC

		CBLOCK 	0x50	;
		Flags			;byte for indicator flag bits
		TempData		;temporary data in main routines 
		TempRxData		;temporary data in Rx buffer routines 
		TempTxData		;temporary data in Tx buffer routines 
		TxStartPtrH		;pointer to start of data in Tx buffer
		TxStartPtrL		;pointer to start of data in Tx buffer
		TxEndPtrH		;pointer to end of data in Tx buffer
		TxEndPtrL		;pointer to end of data in Tx buffer
		RxStartPtrH		;pointer to start of data in Rx buffer
		RxStartPtrL		;pointer to start of data in Rx buffer
		RxEndPtrH		;pointer to end of data in Rx buffer
		RxEndPtrL		;pointer to end of data in Rx buffer
		TxBuffer:TX_BUF_LEN	;Tx buffer for data to transmit
		RxBuffer:RX_BUF_LEN	;Rx buffer for received data
		ENDC



		ORG     0x0050		;place code after reset vector

;------------------------------------
;Read data from the transmit buffer and put into transmit register.

PutData:	
		btfss	Flags,TxBufEmpty ;check if transmit buffer is empty
		bra		PutDat1		;if not then go transmit
		bcf		PIE1,TXIE	;else disable Tx interrupt
		bra		EndLowInt

PutDat1:
		rcall	GetTxBuffer	;get data from transmit buffer
		movwf	TXREG		;and transmit

;------------------------------------
;Get received data and write into receive buffer.

GetData:
		btfsc	RCSTA,OERR	;if overrun error
		bra		ErrOERR		;then go handle error
		btfsc	RCSTA,FERR	;if framing error
		bra		ErrFERR		;then go handle error
		btfsc	Flags,RxBufFull	;if buffer full
		bra		ErrRxOver	;then go handle error
		movf	RCREG,W		;get received data
		rcall	PutRxBuffer	;and put in buffer
		bra		EndLowInt

;error because OERR overrun error bit is set
;can do special error handling here - this code simply clears and continues

ErrOERR:	
		bcf		RCSTA,CREN	;reset the receiver logic
		bsf		RCSTA,CREN	;enable reception again
		movlw	0xFA;	; MIDI RT Start
		rcall	PutTxBuffer	;put data in transmit buffer
		bra		EndLowInt

;error because FERR framing error bit is set
;can do special error handling here - this code simply clears and continues

ErrFERR:	
		movf	RCREG,W		;discard received data that has error
		bcf		RCSTA,CREN	;reset the receiver logic
		bsf		RCSTA,CREN	;enable reception again
		movlw	0xFB;	; MIDI RT Continue
		rcall	PutTxBuffer	;put data in transmit buffer
		bra		EndLowInt

;error because receive buffer is full
;can do special error handling here - this code checks and discards the data

ErrRxOver:	
		movf	RCREG,W		;discard received data
		bsf 	OutsC,0; 	error LED
		movlw	0xFB;	; MIDI RT Start
		rcall	PutTxBuffer	;put data in transmit buffer
		bra		EndLowInt

;------------------------------------
;End of high priority interrupt routine restores context.

EndHighInt:	
		movff	FSR0L_SHADOW,FSR0L	;restore FSR0L register
		movff	FSR0H_SHADOW,FSR0H	;restore FSR0H register
		retfie	FAST			;return and restore context

;----------------------------------------------------------------------------
;Set up serial port and buffers.

SetupSerial:
		movlw	0xc0		;set tris bits for Tx and RX
		iorwf	TRISC,F
		movlw	SPBRG_VAL	;set baud rate
		movwf	SPBRG
		movlw	0x24		;enable transmission and high baud rate
		movwf	TXSTA
		movlw	0x90		;enable serial port and reception
		movwf	RCSTA
		clrf	Flags		;clear all flags

		rcall	InitTxBuffer	;initialize transmit buffer
		rcall	InitRxBuffer	;initialize receive buffer

		movlw	0x30		;enable Tx and Rx interrupts
		movwf	PIE1
		movlw	0x00		;set Rx low and Tx low priority
		movwf	IPR1
		bsf		RCON,IPEN	;enable interrupt priorities
		movlw	0xc0		;enable global high and low ints
		movwf	INTCON
		return

;----------------------------------------------------------------------------
;Circular buffer routines.

;----------------------------------------------------------------------------
;Initialize transmit buffer.

InitTxBuffer:	
		movlw	HIGH TxBuffer	;take high address of transmit buffer
		movwf	TxStartPtrH	;and place in transmit start pointer
		movwf	TxEndPtrH	;and place in transmit end pointer
		movlw	LOW TxBuffer	;take low address of transmit buffer
		movwf	TxStartPtrL	;and place in transmit start pointer
		movwf	TxEndPtrL	;and place in transmit end pointer
		bcf		Flags,TxBufFull	;indicate Tx buffer is not full
		bsf		Flags,TxBufEmpty ;indicate Tx buffer is empty
		return

;----------------------------------------------
;Initialize receive buffer.

InitRxBuffer:	
		movlw	HIGH RxBuffer	;take high address of receive buffer
		movwf	RxStartPtrH	;and place in receive start pointer
		movwf	RxEndPtrH	;and place in receive end pointer
		movlw	LOW RxBuffer	;take low address of receive buffer
		movwf	RxStartPtrL	;and place in receive start pointer
		movwf	RxEndPtrL	;and place in receive end pointer
		bcf		Flags,RxBufFull	;indicate Rx buffer is not full
		bsf		Flags,RxBufEmpty ;indicate Rx buffer is empty
		return

;----------------------------------------------------------------------------
;Add a byte (in WREG) to the end of the transmit buffer.

PutTxBuffer:	
		bcf		PIE1,TXIE	;disable Tx interrupt to avoid conflict
		btfsc	Flags,TxBufFull	;check if transmit buffer full
		bra		ErrTxBufFull	;go handle error if full
		movff	TxEndPtrH,FSR0H	;put EndPointer into FSR0
		movff	TxEndPtrL,FSR0L	;put EndPointer into FSR0
		movwf	POSTINC0 	;copy data to buffer
;test if buffer pointer needs to wrap around to beginning of buffer memory
		movlw	HIGH (TxBuffer+TX_BUF_LEN) ;get last address of buffer
		cpfseq	FSR0H		;and compare with end pointer
		bra		PutTxBuf1	;skip low bytes if high bytes not equal
		movlw	LOW (TxBuffer+TX_BUF_LEN) ;get last address of buffer
		cpfseq	FSR0L		;and compare with end pointer
		bra		PutTxBuf1	;go save new pointer if at end
		lfsr	0,TxBuffer	;point to beginning of buffer if at end
PutTxBuf1:	
		movff	FSR0H,TxEndPtrH	;save new EndPointer high byte
		movff	FSR0L,TxEndPtrL	;save new EndPointer low byte

;test if buffer is full

		movf	TxStartPtrH,W	;get start pointer
		cpfseq	TxEndPtrH	;and compare with end pointer
		bra		PutTxBuf2	;skip low bytes if high bytes not equal
		movf	TxStartPtrL,W	;get start pointer
		cpfseq	TxEndPtrL	;and compare with end pointer
		bra		PutTxBuf2
		bsf		Flags,TxBufFull	;if same then indicate buffer full

PutTxBuf2:	
		bcf		Flags,TxBufEmpty ;Tx buffer cannot be empty
		bsf		PIE1,TXIE	;enable transmit interrupt
	return

;error because attempting to store new data and the buffer is full
;can do special error handling here - this code simply ignores the byte

ErrTxBufFull:	
		bsf	PIE1,TXIE	;enable transmit interrupt
		return 			;no save of data because buffer is full

;----------------------------------------------
;Add a byte (in WREG) to the end of the receive buffer.

PutRxBuffer:
		btfsc	Flags,RxBufFull	;check if receive buffer full
		bra		ErrRxBufFull	;go handle error if full

		movff	RxEndPtrH,FSR0H	;put EndPointer into FSR0
		movff	RxEndPtrL,FSR0L	;put EndPointer into FSR0
		movwf	POSTINC0 	;copy data to buffer

;test if buffer pointer needs to wrap around to beginning of buffer memory

		movlw	HIGH (RxBuffer+RX_BUF_LEN) ;get last address of buffer
		cpfseq	FSR0H		;and compare with end pointer
		bra		PutRxBuf1	;skip low bytes if high bytes not equal
		movlw	LOW (RxBuffer+RX_BUF_LEN) ;get last address of buffer
		cpfseq	FSR0L		;and compare with end pointer
		bra		PutRxBuf1	;go save new pointer if not at end
		lfsr	0,RxBuffer	;point to beginning of buffer if at end

PutRxBuf1:
		movff	FSR0H,RxEndPtrH	;save new EndPointer high byte
		movff	FSR0L,RxEndPtrL	;save new EndPointer low byte
;test if buffer is full
		movf	RxStartPtrH,W	;get start pointer
		cpfseq	RxEndPtrH	;and compare with end pointer
		bra		PutRxBuf2	;skip low bytes if high bytes not equal
		movf	RxStartPtrL,W	;get start pointer
		cpfseq	RxEndPtrL	;and compare with end pointer
		bra		PutRxBuf2
		bsf		Flags,RxBufFull	;if same then indicate buffer full

PutRxBuf2:	
		bcf	Flags,RxBufEmpty ;Rx buffer cannot be empty
		return

;error because attempting to store new data and the buffer is full
;can do special error handling here - this code simply ignores the byte

ErrRxBufFull:	
		return 			;no save of data because buffer is full

;----------------------------------------------
;Remove and return (in WREG) the byte at the start of the transmit buffer.

GetTxBuffer:	
		btfsc	Flags,TxBufEmpty ;check if transmit buffer empty
		bra		ErrTxBufEmpty	;go handle error if empty

		movff	TxStartPtrH,FSR0H ;put StartPointer into FSR0
		movff	TxStartPtrL,FSR0L ;put StartPointer into FSR0
		movff	POSTINC0,TempTxData ;save data from buffer

;test if buffer pointer needs to wrap around to beginning of buffer memory

		movlw	HIGH (TxBuffer+TX_BUF_LEN) ;get last address of buffer
		cpfseq	FSR0H		;and compare with start pointer
		bra		GetTxBuf1	;skip low bytes if high bytes not equal
		movlw	LOW (TxBuffer+TX_BUF_LEN) ;get last address of buffer
		cpfseq	FSR0L		;and compare with start pointer
		bra		GetTxBuf1	;go save new pointer if at end
		lfsr	0,TxBuffer	;point to beginning of buffer if at end

GetTxBuf1:	
		movff	FSR0H,TxStartPtrH ;save new StartPointer value
		movff	FSR0L,TxStartPtrL ;save new StartPointer value

;test if buffer is now empty

		movf	TxEndPtrH,W	;get end pointer
		cpfseq	TxStartPtrH	;and compare with start pointer
		bra		GetTxBuf2	;skip low bytes if high bytes not equal
		movf	TxEndPtrL,W	;get end pointer
		cpfseq	TxStartPtrL	;and compare with start pointer
		bra		GetTxBuf2
		bsf		Flags,TxBufEmpty ;if same then indicate buffer empty

GetTxBuf2:	
		bcf		Flags,TxBufFull ;Tx buffer cannot be full
		movf	TempTxData,W	;restore data from buffer
	return

;error because attempting to read data from an empty buffer
;can do special error handling here - this code simply returns zero

ErrTxBufEmpty:	
	retlw	0		;buffer empty, return zero

;----------------------------------------------
;Remove and return (in WREG) the byte at the start of the receive buffer.

GetRxBuffer:
		bcf		PIE1,RCIE	;disable Rx interrupt to avoid conflict
		btfsc	Flags,RxBufEmpty ;check if receive buffer empty
		bra		ErrRxBufEmpty	;go handle error if empty

		movff	RxStartPtrH,FSR0H ;put StartPointer into FSR0
		movff	RxStartPtrL,FSR0L ;put StartPointer into FSR0
		movff	POSTINC0,TempRxData ;save data from buffer

;test if buffer pointer needs to wrap around to beginning of buffer memory

		movlw	HIGH (RxBuffer+RX_BUF_LEN) ;get last address of buffer
		cpfseq	FSR0H		;and compare with start pointer
		bra		GetRxBuf1	;skip low bytes if high bytes not equal
		movlw	LOW (RxBuffer+RX_BUF_LEN) ;get last address of buffer
		cpfseq	FSR0L		;and compare with start pointer
		bra		GetRxBuf1	;go save new pointer if at end
		lfsr	0,RxBuffer	;point to beginning of buffer if at end

GetRxBuf1:	
		movff	FSR0H,RxStartPtrH ;save new StartPointer value
		movff	FSR0L,RxStartPtrL ;save new StartPointer value

;test if buffer is now empty

		movf	RxEndPtrH,W	;get end pointer
		cpfseq	RxStartPtrH	;and compare with start pointer
		bra		GetRxBuf2	;skip low bytes if high bytes not equal
		movf	RxEndPtrL,W	;get end pointer
		cpfseq	RxStartPtrL	; and compare with start pointer
		bra		GetRxBuf2
		bsf		Flags,RxBufEmpty ;if same then indicate buffer empty

GetRxBuf2:	
		bcf		Flags,RxBufFull ;Rx buffer cannot be full
		movf	TempRxData,W	;restore data from buffer
		bsf		PIE1,RCIE	;enable receive interrupt
	return

;error because attempting to read data from an empty buffer
;can do special error handling here - this code simply returns zero

ErrRxBufEmpty:	
		bsf		PIE1,RCIE	;enable receive interrupt
	retlw	0		;buffer empty, return zero

;----------------------------------------------------------------------------
;Move data from receive buffer to transmit buffer to echo the line back.

CopyRxToTx:
Copy1:
		btfsc	Flags,RxBufEmpty ;check if Rx buffer is empty
		return			;if so then return
		rcall	GetRxBuffer	;get data from receive buffer
		rcall	PutTxBuffer	;put data in transmit buffer
		bra		Copy1
		return

;----------------------------------------------------------------------------
; MIDI Input parser
;
; pseudocode:
;
;  MidiInParser:
;	if (inByte & 0x80)
;		MidiStatus = inByte
;		goto MidiInStatus
;	else
;		goto MidiInData
;	return
;
; MidiInStatus:
;	Look up # databytes
;		1000xxxx	-> 2 (noteOn)
;		1001xxxx	-> 2 (noteOff)
;		1010xxxx	-> 2 (polyKeyPr)
;		1011xxxx	-> 2 (ctrlChange)
;		1100xxxx	-> 1 (pgmChg
;		1101xxxx	-> 1 (chnlPr)
;		1110xxxx	-> 0 (System Common Messages -> ignore)
;		1111xxxx	-> 0 (SysEx)
;
;   store in MidiNumData
;   MidiCurDataBytes=0
;   return
; 
; MidiInData:
;	If (MidiCurData==0)
;		MidiData1=inByte
;	If (MidiCurData==1)
;		MidiData2=inByte
;   MidiCurData++;
;
;	If (MidiCurData==MidiNumData)
;	{
;		If (MidiStatus==NoteOn + channel)
;		   	    ...
;
;		If (MidiStatus==NoteOff + channel)
;	}       		...
;

PollMidiIn:
		btfsc	Flags,RxBufEmpty ;check if Rx buffer is empty
		return			;if so then return
		rcall	GetRxBuffer	;get data from receive buffer
MidiInParser:
		movwf	MidiInByte;
;		rcall	PutTxBuffer;	MIDI thru for test purposes
;	return
		movf	MidiInByte,W;
		btfss	MidiInByte,7;
		bra 	MidiInData;	; jump if it wasn't a status byte
		movwf	MidiByte0;
MidiInStatus:
		clrf	MidiCurData;	
		btfss	MidiByte0,6;
		bra		MidiInStatus2words;
		btfss	MidiByte0,5;
		movlw	1;
		movwf	MidiNumData;
	return
MidiInStatus2words:
		movlw	2;
		movwf	MidiNumData;	
MidiInStatusEnd:
	return

MidiInData:
		tstfsz	MidiCurData;
		bra		MidiInDataM2;
		bra		MidiInDataM1;
MidiInDataM2:
	; 2nd byte
		movff	MidiInByte,MidiByte2;
		clrf	MidiCurData;
		bra		MidiInAction2Bytes;
MidiInDataM1:
		movff	MidiInByte,MidiByte1;
		incf	MidiCurData;
		movf	MidiCurData,W;
		subwf	MidiNumData,W;
		bz		MidiInAction1Byte;
	return;
	
MidiInAction1Byte:
	return;

MidiInAction2Bytes:
		movlw	090h + MidiChannel; 			noteOn
		subwf	MidiByte0,0;
		bz		MidiInNoteOn1;
		movlw	080h + MidiChannel; 			noteOff
		subwf	MidiByte0,0;
		bz		MidiInNoteOff1;
		movlw	0B0h + MidiChannel; 			CtrlChange
		subwf	MidiByte0,0;
		bz		MidiInCtrlChange1;
	return;

; proxies for a far jump
MidiInNoteOn1:
		bra 	MidiInNoteOn

MidiInNoteOff1:
		bra 	MidiInNoteOff

MidiInCtrlChange1:
		bra 	MidiInCtrlChange
